//Brent Cowan Aug. 2008

#include <windows.h>		
#include "Sound.h"

listenerClass::listenerClass()
{
	position.x = 0.0f; position.y = 0.0f; position.z = 0.0f;
	velocity.x = 0.0f; velocity.y = 0.0f; velocity.z = 0.0f;
	lastPos.x  = 0.0f; lastPos.y  = 0.0f; lastPos.z  = 0.0f;
	forward.x  = 0.0f; forward.y  = 0.0f; forward.z = -1.0f; 
    up.x       = 0.0f; up.y       = 1.0f; up.z       = 0.0f;
}

SoundEngine::SoundEngine()
{
	DISTANCEFACTOR = 0.3f;

	//Create a System object and initialize.
    result = FMOD::System_Create(&system);
    ERRCHECK(result);
    
    result = system->getVersion(&version);
    ERRCHECK(result);

    if (version < FMOD_VERSION)
    {
        //printf("Error!  You are using an old version of FMOD %08x.  This program requires %08x\n", version, FMOD_VERSION);
        exit(-1);
    }
    
    result = system->getDriverCaps(0, &caps, 0, 0, &speakermode);
    ERRCHECK(result);

    result = system->setSpeakerMode(speakermode);       /* Set the user selected speaker mode. */
    ERRCHECK(result);

    if (caps & FMOD_CAPS_HARDWARE_EMULATED)             /* The user has the 'Acceleration' slider set to off!  This is really bad for latency!. */
    {                                                   /* You might want to warn the user about this. */
        result = system->setDSPBufferSize(1024, 10);
        ERRCHECK(result);
    }

    result = system->getDriverInfo(0, name, 256, 0);
    ERRCHECK(result);

    if (strstr(name, "SigmaTel"))   /* Sigmatel sound devices crackle for some reason if the format is PCM 16bit.  PCM floating point output seems to solve it. */
    {
        result = system->setSoftwareFormat(48000, FMOD_SOUND_FORMAT_PCMFLOAT, 0,0, FMOD_DSP_RESAMPLER_LINEAR);
        ERRCHECK(result);
    }

    result = system->init(100, FMOD_INIT_NORMAL, 0);
    if (result == FMOD_ERR_OUTPUT_CREATEBUFFER)         /* Ok, the speaker mode selected isn't supported by this soundcard.  Switch it back to stereo... */
    {
        result = system->setSpeakerMode(FMOD_SPEAKERMODE_STEREO);
        ERRCHECK(result);
                
        result = system->init(100, FMOD_INIT_NORMAL, 0);/* ... and re-init. */
        ERRCHECK(result);
    }

    //Set the distance units. (meters/feet etc).
    result = system->set3DSettings(1.0, 1.0f, 1.0f);
    ERRCHECK(result);
}

SoundEngine::~SoundEngine()
{
	//Shut down
    result = system->close();
    ERRCHECK(result);
    result = system->release();
    ERRCHECK(result);
}

void SoundEngine::ERRCHECK(FMOD_RESULT result)
{
    if (result != FMOD_OK)
    {
        //printf("FMOD error! (%d) %s\n", result, FMOD_ErrorString(result));
        exit(-1);
    }
}

void SoundEngine::Update(float ElapsedTime)
{
    // velocity = how far we moved last FRAME (m/f), then time compensate it to SECONDS (m/s).
	listener.velocity.x = (listener.position.x - listener.lastPos.x) * (1.0f / ElapsedTime);
    listener.velocity.y = (listener.position.y - listener.lastPos.y) * (1.0f / ElapsedTime);
    listener.velocity.z = (listener.position.z - listener.lastPos.z) * (1.0f / ElapsedTime);

    // store pos for next time
	listener.lastPos = listener.position;

	FMOD_VECTOR tempPos, tempVel;
	tempPos = listener.position; tempVel = listener.velocity;
	tempPos.x *= DISTANCEFACTOR; tempPos.y *= DISTANCEFACTOR; tempPos.z *= DISTANCEFACTOR;
	tempVel.x *= DISTANCEFACTOR; tempVel.y *= DISTANCEFACTOR; tempVel.z *= DISTANCEFACTOR;

    result = system->set3DListenerAttributes(0, &tempPos, &tempVel, &listener.forward, &listener.up);
    ERRCHECK(result);

	system->update();
}


Sound::Sound()
{
	S           = NULL;
	soundFile   = NULL;
	is3D        = false;
	init        = false;
	minDistance = 10.0f;
	maxDistance = 1000.0f;
	position.x = 0.0f; position.y = 0.0f; position.z = 0.0f; 
	velocity.x = 0.0f; velocity.y = 0.0f; velocity.z = 0.0f;
}

Sound::~Sound()
{
	if(init)
	{
		result = soundFile->release();
		S->ERRCHECK(result);
	}
}

bool Sound::Load(char *FileName, SoundEngine* engine, bool is3D, bool isLooping)
{
	S = engine;
	this->is3D = is3D;

	if(soundFile!=NULL)//release memory if there is already a sound loaded 
	{
		soundFile->release();
		soundFile = NULL;
	}

	if(is3D)
	{
		result = S->system->createSound(FileName, FMOD_3D, 0, &soundFile);
		S->ERRCHECK(result);
		result = soundFile->set3DMinMaxDistance(minDistance * S->DISTANCEFACTOR, maxDistance * S->DISTANCEFACTOR);
		S->ERRCHECK(result);
	}
	else
	{
		result = S->system->createSound(FileName, FMOD_2D, 0, &soundFile);
		S->ERRCHECK(result);
	}

	if(isLooping)
	{
		result = soundFile->setMode(FMOD_LOOP_NORMAL);
		S->ERRCHECK(result);
	}
	else
	{
		result = soundFile->setMode(FMOD_LOOP_OFF);
		S->ERRCHECK(result);
	}

	init = true;

	return init;
}

void Sound::Play()
{
	result = S->system->playSound(FMOD_CHANNEL_FREE, soundFile, true, &channel);
    S->ERRCHECK(result);

	if(is3D)
	{
		FMOD_VECTOR tempPos, tempVel;
		tempPos = position; tempVel = velocity;
		tempPos.x *= S->DISTANCEFACTOR; tempPos.y *= S->DISTANCEFACTOR; tempPos.z *= S->DISTANCEFACTOR;
		tempVel.x *= S->DISTANCEFACTOR; tempVel.y *= S->DISTANCEFACTOR; tempVel.z *= S->DISTANCEFACTOR;
		result = channel->set3DAttributes(&tempPos, &tempVel);
		S->ERRCHECK(result);
	}

	result = channel->setPaused(false);
    S->ERRCHECK(result);	
}

void Sound::Set3dPosition(float x, float y, float z)
{
	position.x = x; position.y = y; position.z = z; 

	if(is3D)
	{
		FMOD_VECTOR tempPos, tempVel;
		tempPos = position; tempVel = velocity;
		tempPos.x *= S->DISTANCEFACTOR; tempPos.y *= S->DISTANCEFACTOR; tempPos.z *= S->DISTANCEFACTOR;
		tempVel.x *= S->DISTANCEFACTOR; tempVel.y *= S->DISTANCEFACTOR; tempVel.z *= S->DISTANCEFACTOR;
		result = channel->set3DAttributes(&tempPos, &tempVel);
		S->ERRCHECK(result);
	}
}

FMOD_VECTOR Sound::Get3dPosition()
{
	return position;
}

void Sound::Set3dVelocity(float x, float y, float z)
{
	velocity.x = x; velocity.y = y; velocity.z = z; 
}

FMOD_VECTOR Sound::Get3dVelocity()
{
	return velocity;
}

void Sound::SetMinMaxDistance(float min, float max)
{
	minDistance = min;
	maxDistance = max;

	bool playing;
	channel->isPlaying(&playing);
	if(playing)
	{
		result = channel->set3DMinMaxDistance(minDistance * S->DISTANCEFACTOR, maxDistance * S->DISTANCEFACTOR);
		S->ERRCHECK(result);
	}
}

void Sound::Stop()
{
	channel->stop();
}